static char query_contents_c_sourceCodeID[]="$Id: @(#)03  1.4.1.1 src/pdweb/wand/intraverse/query_contents/query_contents.c, pdweb.webseal, amweb510, 031011a 03/09/25 13:06:13 $";

/*
 *	query_contents.c
 *
 *	An implementation of the CGI program for determining the contents of
 *	the web space of a third party web server.  Returns multi-line
 *	plain text output in the following format:
 *
 *	content-type: text/plain
 *
 *	return-code
 *	files/directories as applicable, one per line.
 *
 *	This implementation is particular to the WIN/32 platform, although
 *	it could be easily adapted to other platforms not supporting a UNIX
 *	shell environment.
 *
 *	No support currently exists for directory mapping.  This could be added by
 *	looking up the registry entries for a particular web server, and incorporating
 *	this into the directory_list function.
 *
 *  Command Line Format:
 *
 *		query_contents dirlist=dir
 *
 *	History:
 *
 *	09-13-96	created
 */

/* Notes:
 * As a cgi, it's reasonably important that we don't intruduce unwanted security holes
 * So, we don't allow '.'s in paths, we don't parse %xy or Unicode definitions
 * and it won't compile on Unix until you set the webserver document root path
 * Don't GUESS, make SURE you get that right.
 * Policy Director doesn't need this to do cute things, we only look through the 
 * web space from root to leaves so we can check that objects exist in the web space.
 */


/* If it's not WIN32, we assume it's something unixy */
#if defined(WIN32)

#include <windows.h>
#include <stdio.h>
#include <string.h>

#else /* Unix */

#include <stdio.h>
#define MAX_LEN 1024
#include <stdlib.h>
#include <ctype.h>
#include <stddef.h>
#include <glob.h>
#include <string.h>

#if !defined(MAX_PATH)
#define MAX_PATH MAX_LEN
#endif

#endif /* Unix */

/* error codes */
#define EXIT_OK			"100\n"
#define EXIT_USAGE		"101\n"
#define EXIT_NO_OBJ		"102\n"
#define EXIT_NO_DIR		"103\n"


/* local prototypes */
static void parse_arguments( char *src, char *lhs, char *rhs );
static void directory_list( char *object );
static void object_check( char *object );
static char *get_docroot( void );
static char *full_object_path( char *object );
static void fwdslash_to_backslash( char *string );


/*
 *	main
 *		Entry point to the program.
 *
 */
int
main( int argc, char *argv[] )
{
	static char command[MAX_PATH+1], nsobject[MAX_PATH+1];

	/* server needs this info for CGI return */
	printf( "content-type: text/plain\n\n" );


	if ( argc == 2 ) {
	  /* First, trim the input down to reduce problems further on */
	  if(strlen(argv[1]) >= MAX_PATH) argv[1][MAX_PATH] = '\0';
	  parse_arguments( argv[1], command, nsobject );
	}
	else {
		parse_arguments( getenv("QUERY_STRING"), command, nsobject );
	}

	/* branch on command */
	if ( strcmp( command, "dirlist" ) == 0 ) {
#if defined(WIN32)	  
		strncat( nsobject, "\\",(MAX_PATH-strlen(nsobject)) );
#else
		strncat( nsobject, "/",(MAX_PATH-strlen(nsobject)) );
#endif
		/* We check for '.'s in the input here and simply say NO 
		 * PD never uses this.
		*/
		if( strchr(nsobject,'.') != NULL) {
		  printf(EXIT_USAGE);
		  exit(1);
		}		
  
		directory_list( nsobject );
	}
	else {
		printf( EXIT_USAGE );
		return 0;
	}
	return 0 ;
}


/*
 *	parse_arguments
 *		Build up the command and name space object strings.  This function
 *  	will exit if it cannot find these strings properly.
 *
 *	PARAMETERS
 *		src		The argument as passed to the program from the command line or
 *				QUERY_STRING environment variable.
 *		lhs		The lhs result of parsing
 *		rhs		The rhs result of parsing
 *
 *	RETURNS
 *		nothing
 */
static void
parse_arguments( char *src, char *lhs, char *rhs )
{
	char *rhs_orig = rhs; 
	int i;

	if ( src == (char *)NULL ) {
		printf( EXIT_NO_DIR );
		exit(0);
	}

	
	i = 0;
	while ( *src && *src != '=' && (i < MAX_PATH) ) {
		*lhs++ = *src++;
		i++;
	}
	*lhs = '\0';

	if ( *src == '\0' ) {
		/* we reached the end of the argument before the = */
		printf( EXIT_NO_DIR );
		exit(0);
	}
	
	src++;	/* to skip the = */
        i = 0;
	while ( *src  && (i < MAX_PATH) ) {
		*rhs++ = *src++;
                i++;
	}
        *rhs = '\0';

#if defined(WIN32)
	/* now convert all to back slashes for the directory lookup */
	fwdslash_to_backslash( rhs_orig );
#endif
}


/*
 *	directory_list
 *		Lists objects in the directory specified.  The directory is
 *  	relative (or absolute to the "document root").
 *
 *	PARAMETERS
 *		object		A path relative to the document root.  This is converted into an
 *					absolute path for use within this function.
 *
 *	RETURNS
 *		none
 */
#if defined(WIN32)
static void 
directory_list( char *object )
{
	char				*fop = full_object_path( object );
	WIN32_FIND_DATA		win32fData;
	HANDLE				win32fHandle;

	strncat( fop, "\\*.*", (MAX_PATH-strlen(fop)));
	win32fHandle = FindFirstFile( fop, &win32fData );
	if ( win32fHandle != INVALID_HANDLE_VALUE ) {
		printf( EXIT_OK );
	}
	else {
		printf( EXIT_NO_DIR );
		exit (0);
	}

	while ( 1 ) {
		if ( strcmp( win32fData.cFileName, "." ) &&
			 strcmp( win32fData.cFileName, ".." ) ) {
			printf( win32fData.cFileName );
			if ( win32fData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) {
				/* directories are signified with the double slash */
				printf( "//" );
			}
			printf( "\n" );
		}
		if ( !FindNextFile( win32fHandle, &win32fData ) ) {
			break;
		}
	}
	FindClose( win32fHandle );
}

#else /* Unixy style */
static void 
directory_list( char *object )
{
	char *fop = full_object_path( object );
	glob_t globby;	
	int err,i;
	
	strncat( fop, "*",(MAX_PATH-strlen(fop)) );
	
	err = glob(fop,GLOB_MARK,NULL,&globby);
	
	
	if ( err == 0 ) {
		printf( EXIT_OK );
	}
	else {
		printf( EXIT_NO_DIR );
		exit (0);
	}
	for( i = 0; i < globby.gl_pathc; i++) {
		char *p;
		p = globby.gl_pathv[i];	
		if( *p == '/') {		
			printf("%s",p+1);
		}
		if( *(p + strlen(p)-1) == '/') printf("/");
		printf("\n");		
	} 
	globfree(&globby);

}
#endif

/*
 *	get_docroot
 *		Returns a pointer to a string that is the document root for the
 *		web space.  We obtain this from a plain text configuration file on
 *  	the web server, in the same directory as the query_contents script.
 *		We assume that this file will be in "the current directory" when
 *		this program runs.  The name of this file is query_contents.cfg as
 *		the format is as follows:
 *
 *			[server]
 *			docroot=
 *
 *		It may be desirable to obtain this directly from the registry entry for
 *		the web server in question.  Of course, this would then make the code
 *		web server specific.
 *
 *	PARAMETERS
 *		none
 *
 *	RETURNS
 *		The document root obtained from the configuration file.
 */
static char *
get_docroot( void )
{
#if !defined(WIN32)
#error   "You MUST set the documument root directory for the webserver, and comment out this line"
        static char docroot[MAX_PATH+1] = {"/usr/local/www"};
       	 
#else
	static char docroot[MAX_PATH+1];

	GetPrivateProfileString(	"server", 
								"docroot", 
								"",
								docroot,
								MAX_PATH-1,
								"query_contents.cfg" );
	printf("docroot=[%s]\n",docroot ? docroot: "docroot is NULL");
	if ( docroot[0] == '\0' ) {
		printf( EXIT_USAGE );
		exit(0);
	}
#endif
	return docroot;
}


/*
 *	full_object_path
 *		Takes a pointer which is the object string and returns a pointer to
 *		the full path of the object, in the WIN/32 filespace.  This function
 *		takes into consideration all of the slash issues along the way.
 *
 *	PARAMETERS
 *		object		Name of the object string, relative to the document root.
 *
 *	RETURNS
 *		The full path of the object on the local file system.
 */
#if defined(WIN32)
static char *
full_object_path( char *object )
{
	static char full_obj_path[MAX_PATH+1];

	strncpy( full_obj_path, get_docroot(),MAX_PATH );
	if ( full_obj_path[strlen(full_obj_path)-1] == '\\' ) {
		if ( object[0] == '\\' ) {
			strncat( full_obj_path, &object[1], (MAX_PATH - strlen((char*)full_obj_path)) );
		}
		else {
			strncat( full_obj_path, object, (MAX_PATH - strlen((char*)full_obj_path)) );
		}
	}
	else {
		if ( object[0] == '\\' ) {
			strncat( full_obj_path, object, (MAX_PATH - strlen((char*)full_obj_path)) );
		}
		else {
			strncat( full_obj_path, "\\", (MAX_PATH - strlen((char*)full_obj_path)) );
			strncat( full_obj_path, object, (MAX_PATH - strlen((char*)full_obj_path)) );
		}
	}
	return full_obj_path;
}
#else
static char *
full_object_path( char *object )
{
	static char full_obj_path[MAX_PATH+1];

	strncpy( full_obj_path, get_docroot(),MAX_PATH );
printf("full_obj_path: %s\n", full_obj_path);
	if ( full_obj_path[strlen(full_obj_path)-1] == '/' ) {
		if ( object[0] == '/' ) {
			strncat( full_obj_path, &object[1],(MAX_PATH - strlen(full_obj_path)) );
		}
		else {
			strncat( full_obj_path, object,(MAX_PATH-strlen(full_obj_path)) );
		}
	}
	else {
		if ( object[0] == '/' ) {
			strncat( full_obj_path, object,(MAX_PATH-strlen(full_obj_path)) );
		}
		else {
			strncat( full_obj_path, "/",(MAX_PATH-strlen(full_obj_path)) );
			strncat( full_obj_path, object,(MAX_PATH-strlen(full_obj_path)) );
		}
	}
	return full_obj_path;
}
#endif

/*
 *	fwdslash_to_backslash
 *		Converts from UNIX-style to MS-DOS style directory separators.
 *
 *	PARAMETERS
 *		string		String to convert.  Note that the conversion is done inline.
 *
 */
static void 
fwdslash_to_backslash( char *string )
{
	char *c = string;

	if ( !c ) {
		return;
	}

	while ( *c ) {
		if ( *c == '/' ) {
			*c = '\\';
		}
		c++;
	}
}
